/**
  * Header file to support JBurg whilst emitting C++
  */
#include <string>
#include <stdexcept>
#include <vector>

// strange as it looks, this does work (at least on my platform, VS.NET 2003)
// and I think it should work on any platform
#define MAX_INT_VALUE (-1^(1<<((sizeof(int)*8)-1)))
// and the inverse (might be useful!)
#define MIN_INT_VALUE (MAX_INT_VALUE+1)


template<class T>
class JBurgAnnotation
{

public:
	class JBurgSubgoal
	{
	public:
		JBurgSubgoal(JBurgAnnotation<T>* node, const int subgoal_state):
			m_node(node)
			, m_subgoal_state(subgoal_state)
		{
		}

	public:

		/// @return the node to be reduced.
	    JBurgAnnotation<T>* getNode() { return m_node; }

		/// @return the state to which the node must be reduced.
		const int getGoalState() const { return m_subgoal_state; }

	private:
		JBurgAnnotation<T>* m_node;
		int m_subgoal_state;
	};

private:
	///  The node being labelled/reduced.
	T*					m_node;

	///  Denormalized data for ease of reference.
	///  FIXME: Could be static?
	int					nRules;

	///  Per-goal-state cost values.
	int					*cost;

	///  Per-goal-state cost values.
	int					*rule;

	int                 *narySubgoal;

	///  Subgoals to be reduced before reducing this
	///  annotation's node; these are non-terminal
	///  parameters to the pattern matching rule.
	std::vector< std::vector<JBurgSubgoal> > m_subgoals;

	///  Antecedent states of the a given state.
	std::vector<int> m_antecedent_states;

private:
	/*
	 * Subtree structure.
	 */
	std::vector< JBurgAnnotation<T>* > m_children;

/* c'tors & d'tor */
public:
	JBurgAnnotation(T* newNode, int nRules)
	{
		m_node = newNode;
		
		// initialise rules array
		rule = new int[nRules];
		memset(rule, 0, sizeof(int)*nRules);

		narySubgoal = new int[nRules];
		memset(narySubgoal, 0, sizeof(int)*nRules);
		
		// init cost array
		cost = new int[nRules];
		for(int x=0; x<nRules; ++x)
		{
			cost[x] = MAX_INT_VALUE;
		}
		
		// initial rule for all sub-goals is undefined.
		
		// keep a copy of the array sizes, we don't know them otherwise! (unlike java we can't do array.length)
		this->nRules = nRules;
	}

	~JBurgAnnotation()
	{
		// assume that m_node is a reference & is deleted elsewhere
		if(rule) delete[] rule;
		if(cost) delete[] cost;
		
		for ( size_t i = 0; i < m_children.size(); i++ )
		{
			delete m_children[i];
		}
	}


protected:
	// Undefine the default constructor
	JBurgAnnotation();


/* methods */
public:
	int				getOperator()					{ return m_node->getType(); }
	T*				getNode()						{ return m_node; }
	std::string&	toString()						{ return m_node->getText(); }
	int				getCost(int goalState)			{ return cost[goalState]; }
	int				getRule(int goalState)			{ return rule[goalState]; }
	int             getNarySubgoal(int goalState)   { return narySubgoal[goalState]; }
	bool            isNary(int goalState)           { return getNarySubgoal(goalState) != 0; }
	
	void reset(int goalState, int cost, int rule) {
		this->cost[goalState] = cost;
		this->rule[goalState] = rule;

		if ( m_subgoals.size() > goalState)
			m_subgoals[goalState].clear();

		this->narySubgoal[goalState] = 0;

	}

	const bool hasAntecedent( int goalState )
	{
		return m_antecedent_states.size() > goalState && m_antecedent_states[goalState] != 0;
	}

	const int getAntecedent( int goalState )
	{
		return m_antecedent_states[goalState];
	}

	void recordAntecedent(const int goal_state, const int antecedent_state)
	{
		if ( m_antecedent_states.size() <= goal_state )
		{
			m_antecedent_states.resize(nRules + 1);
		}
		m_antecedent_states[goal_state] = antecedent_state;
	}

	void addSubgoal(int goalState, JBurgAnnotation<T>* node, int subGoal) 
	{
		if ( m_subgoals.size() == 0 )
		{
			m_subgoals.resize(nRules + 1);
		}

		m_subgoals[goalState].push_back(JBurgSubgoal(node, subGoal));

	}

	size_t getSubgoalsSize(int goalState)
   	{
		if ( m_subgoals.empty())
		{
			return 0;
		}
		else
		{
			return m_subgoals[goalState].size();
		}
	}

	const std::vector<JBurgSubgoal>& getSubgoals( int goalState)
	{
		if ( m_subgoals.empty())
		{
			throw new std::runtime_error("No subgoal records." );
		}

		return m_subgoals[goalState];
	}

	void setNarySubgoal(int goalState, int subgoalState)
	{
		narySubgoal[goalState] = subgoalState;
	}

	/**
	 *  @return the number of children of this node.
	 */
	size_t getArity()
	{
		return m_children.size();
	}

	/**
	 *  @return the child at the specified index.
	 *  @throw std::runtime_error if the child index is out of bounds.
	 */
	JBurgAnnotation<T>* getNthChild(const size_t idx)
	{
		if ( idx < getArity() )
		{
			return m_children[idx];
		}
		else
		{
			throw new std::runtime_error("getNthChild() index out of bounds.");
		}
	}

	/**
	 *  Add a new child node.
	 *  @post this node assumes ownership of the child node.
	 */
	void addChild(JBurgAnnotation<T>* new_child)
	{
		if ( new_child)
			m_children.push_back(new_child);
	}
	
};
