package jburg.burg;

//  BURM specification is parsed into ANTLR2 AST objects.
import antlr.collections.AST;

//  The code emitter interface definition.
import jburg.burg.emitlangs.EmitLang;

//  The factory that selects and creates an emitter.
import jburg.burg.emitlangs.JBurgEmitterFactory;

//  The i-node adapter interface definition.
import jburg.burg.inode.InodeAdapter;
import jburg.burg.inode.InodeAdapter2;
//  Interface for i-node adapters that need
//  to emit support routines.
import jburg.burg.inode.InodeAuxiliarySupport;

//  The factory that selects and creates an I-node adapter.
import jburg.burg.inode.InodeAdapterFactory;

//  Token types from ANTLR.
import jburg.parser.JBurgTokenTypes;

//  Version file, generated by build.
import jburg.generator.version.JBurgVersion;


import java.io.PrintStream;

import java.lang.reflect.Modifier;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;


@SuppressWarnings("nls")
public class JBurgGenerator implements JBurgTokenTypes
{
    /**
     *  The JBurg specification's rules are categorized as:
     *  <ul>
     *  <li> Pattern  (e.g., integer = PLUS(integer, integer) )
     *  <li> Simple Transformational   (e.g., registerOperand  = integer)
     *  <li> Complex Transformational  (e.g., integer = string { code to convert string to integer} )
     *  </ul>
     *
     *  Simple transformational rules allow a subgoal to satisfy
     *  other subgoals without additional processing; complex
     *  transformational rules allow one subgoal to feed additional
     *  processing that can satisfy other subgoals (at added cost).
     *
     *  Syntactically, a simple transformational rule is distinguished 
	 *  by its lack of a block of code.
	 *  Simple transformational rules are also known as "chain rules"
	 *  in simpler BURGs that don't support complex transformational
	 *  rules.  This longer name was adopted to highlight that difference,
	 *  but a simple transformational rule is, in fact, a chain rule
	 *  by a different name.
     */

	/**
     *   The table of all transformation rules; keyed
     *   by the rules' necessary state, e.g., integer
	 *   in the rule number = integer).
     */
    Map<String, Vector<JBurgRule>> transformationRules = new HashMap<String, Vector<JBurgRule>>();

	/**
	 *  The table of all pattern rules, keyed
	 *  by the top-level operator of the pattern,
	 *  e.g., PLUS in the pattern integer = PLUS(integer, integer).
	 */
    Map<String, Vector<JBurgRule>> patternRules = new HashMap<String, Vector<JBurgRule>>();

    /**
     *  The goal states' names become symbolic constants
     *  in the generated reducer.
     */
    Set<String> all_goal_states = new HashSet<String>();

    /**
     *  Closure sets are organized by their goal state; each state's
     *  closure set contains an entry for every transitive closure
     *  path from any other non-terminal state to the target state.
     */
    Map<String, Vector<ClosureRecord>> closureSets = new HashMap<String, Vector<ClosureRecord>>();

    /**  Action code fragments.  */
    Vector<JBurgReduceAction> reduceActions = new Vector<JBurgReduceAction>();

    /**  Cost function fragments.  */
    Vector<AST> cost_functions = new Vector<AST>();

    /**  The names of any interfaces that the generated BURM implements.  */
    Vector<String> m_InterfaceNames = new Vector<String>();
    
    /** blocks of code to be added into the class verbatim */
    Vector<String> m_inclassCode = new Vector<String>();

    /** The package name of the generated reducer. */
    String packageName = null;

    /**  Header code, copied as-is into the reducer. */
    String headerBlock = null;

    /**
     *  The name of the i-node class that's being labeled and reduced.
     */
    String iNodeClass = null;

	/** 
	 *  I-node adapter class name.  The adapter is instantiated by name.
	 *  @note: specified as an alternative to the I-node class' name,
	 *  which selects a builtin adapter.
	 */
	String iNodeAdapterClass;

    /**
     *  The return types for specific states.
     */
    Map<String, String> returnTypeTable = new HashMap<String, String>();

    /**
     *  The default return type of the reducer functions.
     *  If this is defaulted to null, the generated reducer will return nodes of the iNodeClass.
     */
    String defaultReturnType = null;

    /**
     *  Table of properties to be added to the BURM.
     *  Each one is a name/type pair; the BURM gets
     *  a private member and get/set methods.
     */
    Map<String, String> burm_properties = new HashMap<String, String>();

    /**
     *  Simple transformations all return their input node, so they can all use the same rule.
     *  They're assigned this dummy rule; it will be put into the rule table as rule #1, but
     *  the generated BURM hardcodes special-case logic for rule 1, so it's never called.
     */
    JBurgReduceAction simpleTransformationAction = addAction("dummy SimpleTransformation rule", null);

    /**  Include debugging code in the generated reducer?  */
    boolean m_debugMode;
    
    /** 
     * If the pattern-matcher generator fails, dump its annotation tree here.
     * Note: this is only enabled (or useful) when debugging JBurg's own BURM. 
     */
    String patternMatcherDumpFile = null;

    /**  Caller should set this, it will be set to default values if not provided. */
    Logger logger = null;
    
    /**
     * Name for the language to emit code in (default is assumed to be java (=emitlang.EmitJava)
     */
    String emitLanguageName = null;
    
    /** Code emitter to use (selected using the language name) */
	EmitLang m_codeEmitter = null;

	/** I-node adapter to use */
	InodeAdapter m_adapter;
	
	/** 
	 *  If the InodeAdapter is also an InodeAdapter2,
	 *  keep a reference to it. 
	 */
	InodeAdapter2 adapter2;

	/**  Cumulative error count. */
	int m_errCount = 0;

	/**  Name of the initial parameter to the label routine. */
	private static String m_initalParamName = "to_be_labelled";
	
	/** Name of the node in the reducer */
	private String reducerNodeName = "__p";
	
	/** Name of the stack of reduced values */
	private String reducedValuesName = "__reducedValues";

    /**  Default error handler.  null means use hard-coded default, i.e., throw an exception. */
    String defaultErrorHandler = null;

    /**  Prologue snippets mapped to the corresponding rule number. */
    Map<Integer,String> prologueBlocks = new TreeMap<Integer,String>();

    /** 
     *  All operators known to any pattern accumulate in this set
     *  so that the BURG can generate operator-specific getNthChild()
     *  routines.
     */
    Set<String> allOperators = new TreeSet<String>();
    
    /**
     *  The type of the INodes' opcode.  Defaults to int but
     *  can be an enum for maintainability.
     */
    String opcodeType = "int";

    /**
     *  NamedPattern holds a named pattern, and keeps a table of the reductions that refer to it.
     *  (See defect 3063143 for problems with multiple reductions sharing a pattern.)
     */
    class NamedPattern
    {
        String patternName;
        AST pattern = null;
        Vector<AST> reductions = new Vector<AST>();

        NamedPattern(String name)
        {
            this.patternName = name;
        }
    }

    @SuppressWarnings("serial")
    class PatternMap extends  TreeMap<String,NamedPattern>
    {
        NamedPattern getPattern(String key)
        {
            if ( !super.containsKey(key) )
                put(key, new NamedPattern(key));
            return super.get(key);
        }

        void addPattern(String key, AST pattern)
        {
            NamedPattern p = getPattern(key);
            p.pattern = pattern;
        }

        void addReduction(String key, AST reduction)
        {
            NamedPattern p = getPattern(key);
            p.reductions.add(reduction);
        }
    }

    PatternMap namedPatterns = new  PatternMap();

    /**
      *  @param root - the root of the AST generated by parsing the .jburg specification.
      */
    public JBurgGenerator(AST root, Logger log) throws Exception
    {
        this.logger = log;

        //  Walk over the children of the root, each of which 
		//  is a self-contained syntactical construct, and
        //  find an appropriate action for each.
        for (AST currentNode = root; currentNode != null;
                currentNode = currentNode.getNextSibling()) 
		{
            switch (currentNode.getType()) 
			{
            case COST_FUNCTION:
                cost_functions.add(currentNode);
                break;

            case HEADER_DECLARATION:

				if ( null == this.headerBlock )
				{
					this.headerBlock = getCodeBlock(currentNode);
				}
				else
				{
                    throw new IllegalArgumentException("The class header may only be specified once.");
				}

                break;
                
            case INCLASS_DECLARATION:
				{
            		String ictxt = getCodeBlock(currentNode);

            		// strip the first & last brackets as they are part of our
            		// syntax, not part of the output semantics.
            		
            		int startchar = ictxt.indexOf('{') + 1;
            		int lentoend = ictxt.lastIndexOf('}');
            		ictxt = ictxt.substring( startchar, lentoend-startchar );
            		
            		m_inclassCode.add( ictxt );
            	}
				
				break;

            case INODE_ADAPTER_DECLARATION:

                if (this.iNodeAdapterClass == null)
				{
					this.iNodeAdapterClass = getIdentifierText(currentNode.getFirstChild());
				}
				else
				{
                    throw new IllegalArgumentException("INodeAdapter may only be specified once.");
				}
				break;

            case INODE_TYPE_DECLARATION:

                if (this.iNodeClass == null)
				{
                    this.iNodeClass = getIdentifierText(currentNode.getFirstChild());
				}
                else
				{
                    throw new IllegalArgumentException("INodeType may only be specified once.");
				}

                break;
                
            case LANGUAGE_DECLARATION:

            	if ( null == this.emitLanguageName )
				{
            		this.emitLanguageName = currentNode.getFirstChild().getText();
				}
            	else
				{
            		throw new IllegalArgumentException("Language may only be specified once.");
				}

            	break;

            case IMPLEMENTS_INTERFACE_SPECIFICATION:
                m_InterfaceNames.addElement(getIdentifierText( currentNode.getFirstChild()) );

                break;

            case PACKAGE_SPECIFICATION:
                if ( null == this.packageName )
				{
					this.packageName = getIdentifierText(currentNode.getFirstChild());
				}
				else
				{
            		throw new IllegalArgumentException("package may only be specified once.");
				}

                break;

            case PROPERTY_SPECIFICATION: 
				{
					String propertyType = getIdentifierText(currentNode.getFirstChild());
					String propertyName = currentNode.getFirstChild().getNextSibling().getText();
					burm_properties.put(propertyName, propertyType);

					break;
				}

            case RETURN_DECLARATION:

                if (this.defaultReturnType == null)
                    this.defaultReturnType = getIdentifierText(currentNode.getFirstChild());
                else
                    throw new IllegalArgumentException( "ReturnType may only be specified once.");

                break;

            case PATTERN_RULE:
                addPatternRule(currentNode);

                break;

            case SIMPLE_TRANSFORMATION_RULE:
                addSimpleTransformationRule(currentNode);

                break;

            case TRANSFORMATION_RULE:
                addComplexTransformationRule(currentNode);

                break;

            case TYPED_RETURN_DECLARATION:

                String stateName = currentNode.getFirstChild().getText();
                String returnType = getIdentifierText(currentNode.getFirstChild()
                                                                .getNextSibling());

				//  Put the return declaration in the table, but only once per state.
                Object typeCollision = this.returnTypeTable.put(stateName, returnType);
				if ( null != typeCollision )
				{
                    throw new IllegalArgumentException(
						"A state may only specify one ReturnType."
					);
				}

                break;

            case PATTERN_DECLARATION:
                {
                    String pattern_name = currentNode.getFirstChild().getText();
                    namedPatterns.addPattern(pattern_name, currentNode);
                }
                break;

            case REDUCTION_DECLARATION:
                {
                    String pattern_name = currentNode.getFirstChild().getNextSibling().getText();
                    namedPatterns.addReduction(pattern_name, currentNode);
                }
                break;
            case DEFAULT_ERROR_HANDLER:
                this.defaultErrorHandler = getCodeBlock(currentNode);
                break;
                
            case OPCODE_TYPE:
                this.opcodeType = currentNode.getFirstChild().getText();
                System.out.println("Opcode type: " + this.opcodeType);
                break;

            default:
                throw new IllegalArgumentException("Unknown specification AST type " +
            											String.valueOf(currentNode.getType()));
            }
        }

        //  Set the language emitter.
        m_codeEmitter = JBurgEmitterFactory.getEmitter(emitLanguageName, getLogger());

        if ( m_codeEmitter == null )
		{
			throw new IllegalStateException("Unknown language specified: \""+ emitLanguageName +"\"");
        }

		if ( iNodeClass != null )
		{
			if ( iNodeAdapterClass != null )
			{
				try
				{
					m_adapter = (InodeAdapter)Class.forName(iNodeAdapterClass).newInstance();
				}
				catch ( Exception ex )
				{
					ex.printStackTrace();
					throw new IllegalArgumentException ("Unable to instantiate i-node adapter " + iNodeAdapterClass );
				}
			}
			else
			{
				m_adapter = InodeAdapterFactory.getAdapter(iNodeClass);
			}
			
			m_codeEmitter.setINodeType(this.iNodeClass);

			if ( m_adapter != null )
			{
				logger.info("Using i-node adapter " + m_adapter.getClass().getName() );
			}
			else
			{
				getLogger().warning("using default i-node adapter, no adapter matches " + iNodeClass );
				m_adapter = new jburg.burg.inode.DefaultAdapter();
			}
			
			//  See if the adapter is also an InodeAdapter2 implementation.
			this.adapter2 = (m_adapter instanceof InodeAdapter2)? (InodeAdapter2)m_adapter : null;
		}
		else
		{
			throw new IllegalStateException("You must specify the i-node type.");
		}

		m_codeEmitter.setInodeAdapter(m_adapter);

        //  Default return type is the same as the INode class.
        if (this.defaultReturnType == null)
		{
            this.defaultReturnType = this.iNodeClass;
		}

        //  Mutate pattern/reduction decl pairs into pattern rules.
        for ( NamedPattern np: namedPatterns.values() )
        {
            if ( np.pattern != null && np.reductions.size() > 0 )
            {
                AST named_pattern = np.pattern.getFirstChild().getNextSibling().getFirstChild();

                for ( AST reduction: np.reductions )
                {    
                    //  Splice the pattern into the reduction AST.
                
                    AST nt_state = reduction.getFirstChild();
                    AST pattern_name = nt_state.getNextSibling();
                    AST cost_decl = pattern_name.getNextSibling();
                    
                    //  Create a new holder for the pattern
                    //  so the original AST isn't mutated.
                    AST pattern_holder = new antlr.CommonAST();
                    pattern_holder.setFirstChild(named_pattern);
                    nt_state.setNextSibling(pattern_holder);
                    pattern_holder.setNextSibling(cost_decl);
    
                    //  Give the composite AST the appropriate type
                    //  for its pattern.
                    reduction.setType(PATTERN_RULE);
    
                    addPatternRule(reduction);
                }
            }
            else if ( np.pattern != null )
            {
                getLogger().warning("pattern " + np.patternName + " has no reduction - ignored.");
            }
            else if ( np.reductions.size() > 0 )
            {
                throw new IllegalStateException("Reduction " + np.patternName + " has no associated pattern.");
            }
        }

        //  Compute the closure sets.
        computeClosureSets();
    }
    
    void splicePattern(AST reduction, AST pattern)
    {
        
    }

    /**
     *  Add a reduce action to the action list.
	 *  Actions are keyed by number in the generated BURM.
     */
    private JBurgReduceAction addAction(String strAction, String strState)
    {
        JBurgReduceAction newAction = new JBurgReduceAction(strAction);
        this.reduceActions.add(newAction);

        //  The first index is used as the default action for
        //  simple transformation rules.
        newAction.setIndex(this.reduceActions.size());
        newAction.setState(strState);

        return newAction;
    }

    private JBurgReduceAction decodeReduction(AST parent, String goal_state)
    {
        AST reduction = getASTByType(parent, REDUCTION_ACTION);
        AST action_ast = reduction.getFirstChild();

        JBurgReduceAction action = addAction(action_ast.getText(), goal_state);

        if ( action_ast.getNextSibling() != null )
        {
            String prologue_code = action_ast.getNextSibling().getFirstChild().getText();
            this.prologueBlocks.put(action.m_index, prologue_code);
        }

        return action;
    }

    /**
     *  Add a complex transformation rule to its table,
     *  and record the rule's reduction action.
     */
    private void addComplexTransformationRule(AST newRule)
        throws Exception
    {
        JBurgRule n = addNamedRule(this.transformationRules, newRule);
        this.all_goal_states.add( n.getGoalState() );

        //  Prepare the rule's reduce action.
        JBurgReduceAction action = decodeReduction(newRule, n.getGoalState());

        n.setReduceAction(action);

        //  Add the goal state by name as an alias action routine "parameter."
        //  The "parameter" is popped off the reduced values stack into a local variable.
        action.addParameter(
            newRule.getFirstChild().getNextSibling().getText(),
            newRule.getFirstChild().getNextSibling().getText(),
            ParameterDescriptor.FIXED_ARITY
        );
    }

    /**
     *  Add a simple transformation rule to its table.
     */
    private void addSimpleTransformationRule(AST newRule)
    {
        JBurgRule n = addNamedRule(this.transformationRules, newRule);
        this.all_goal_states.add( n.getGoalState() );
        n.setReduceAction(this.simpleTransformationAction);
    }

    /**
     *  Wrap a rule in an I-node, and add it to the list of rules
     *  that produce a particular subgoal.
     */
    private JBurgRule addNamedRule(Map<String, Vector<JBurgRule>> ruleTable, AST newAST)
    {
        JBurgRule newRule = new JBurgRule(newAST);

        String strRuleName = newRule.getOperator();

        Vector<JBurgRule> vRules = ruleTable.get(strRuleName);

        if (vRules == null) {
            vRules = new Vector<JBurgRule>();
            ruleTable.put(strRuleName, vRules);
        }

        vRules.add(newRule);

        return newRule;
    }

    /**
     *  Wrap a pattern rule in an I-node, and add it to pattern rule table.
     */
    private void addPatternRule(AST newRule) throws Exception
    {
        JBurgRule n = addNamedRule(this.patternRules, newRule);
        this.all_goal_states.add( n.getGoalState() );

        JBurgReduceAction newAction = decodeReduction(newRule, n.getGoalState());

        n.setReduceAction(newAction);
    }


    /**
     *  Compute the closure set of a rule:  the set of rules that
     *  that can transform this rule's goal to satsify another goal.
	 *  @note computeClosure is public so that it can be applied
	 *    by JBurgUtilities.
     */
    public void computeClosure(JBurgRule n) throws Exception
    {
		//  Get the list of transformation rules that can use this rule's goal state.
        Vector<JBurgRule> vClosureCandidate = this.transformationRules.get(n.getGoalState());

        if (vClosureCandidate != null) 
		{
            //  Found some closures:
            //  create or fetch this target state's closure set.
            Vector<ClosureRecord> vClosureSet = this.closureSets.get(n.getGoalState());

            if (vClosureSet == null) 
			{
                vClosureSet = new Vector<ClosureRecord>();
                this.closureSets.put(n.getGoalState(), vClosureSet);
            }

            for (JBurgRule nPrime:vClosureCandidate ) 
			{

                //  Add this closure to the closure set.
                ClosureRecord newClosure = new ClosureRecord(nPrime);

                if (!vClosureSet.contains(newClosure)) 
				{
                    vClosureSet.add(newClosure);
                }
            }
        }
    }

    /**
     *  Scan all the rules and compute their closure sets.
     */
    private void computeClosureSets() throws Exception
    {
        for (Vector<JBurgRule> patterns : this.patternRules.values() ) 
		{
            try 
			{
                JBurgUtilities.applyToAll( this, patterns, "computeClosure", JBurgRule.class);
            }
			catch (Exception e)
			{
                e.printStackTrace();
            }
        }

        for (Vector<JBurgRule> transformations : this.transformationRules.values() )
		{
            try
			{
                JBurgUtilities.applyToAll( this, transformations, "computeClosure", JBurgRule.class);
            }
			catch (Exception e)
			{
                e.printStackTrace();
            }
        }
    }

    /**
     *  Emit the BURG specification's resultant BURM to an output stream.
     */
    public int generate(String strClassName, PrintStream output)
    {
        try
		{
            m_codeEmitter.setOpcodeType(this.opcodeType);
        	emitHeader(strClassName, output);
        	m_codeEmitter.emitInclass(iNodeClass, m_inclassCode, output);

        	emitNTConstants(all_goal_states, output);        	

			if ( m_adapter instanceof InodeAuxiliarySupport )
			{
				((InodeAuxiliarySupport)m_adapter).emitAuxiliarySupport(m_codeEmitter, output);
			}

            emitLabelFunction(iNodeClass, output);
            emitComputeCostMatrixFunction(patternRules, iNodeClass, output);
            emitClosures(closureSets, iNodeClass, output);
            emitActions(reduceActions, iNodeClass, output);
            emitCostFns(cost_functions, iNodeClass, output);
            
            //  If there's an InodeAdapter2 present,
            //  emit the getNthChild routine that
            //  implements the adapter's node-handling logic.
            if ( this.adapter2 != null)
                emitGetNthChild(output);

            m_codeEmitter.emitTrailer(
				strClassName,
				iNodeClass,
				all_goal_states,
				burm_properties,
				m_debugMode,
                this.defaultErrorHandler,
                this.prologueBlocks,
				output
			);
        }
		catch (Exception e)
		{
            e.printStackTrace();
			m_errCount++;
        }

		return m_errCount;
    }

    /**
     *  Get the payload of an identifier, with error checking.
     *  @return the identifier's text.
     */
    private String getIdentifierText(AST p)
    {
		if ( p.getType() != IDENTIFIER )
		{
			throw new IllegalStateException ( "Expected IDENTIFIER, found " + p.toStringTree() );
		}

        return p.getText();
    }

    /**
     *  Does this production have a closure set?
     *  @return true if n has a closure set.
     *  @param n - the production of interest.
     */
    public boolean hasClosure(JBurgProduction n)
    {
        return this.closureSets.get(n.getGoalState()) != null;
    }

    /**
      *  Emit the file/class header information.
      */
    private void emitHeader(String strClassName, PrintStream output)
    {
        output.print( m_codeEmitter.genComment(" Generated " + new java.util.Date().toString() + " by JBurg version " + JBurgVersion.version  ));
        output.print( m_codeEmitter.genBeginLine() );
        m_codeEmitter.emitHeader(strClassName, this.packageName, this.headerBlock, m_InterfaceNames, this.m_debugMode, output);
    }

	/**
	 *  @return the return type for a specific state.
	 */
	public String getReturnType(String stateName)
	{
		Object result = returnTypeTable.get(stateName);

		if (result == null)
			result = defaultReturnType;

		return result.toString();
	}
	
    /**
     *  setDebugMode controls the level of debugging code that will be generated in the reducer.
     */
    public void setDebugMode(boolean bDebugMode)
    {
        m_debugMode = bDebugMode;
    }
    
	/**
	 *  @return the text of the parent AST's code block, which is represented as a BLOCK token.
	 */
	private String getCodeBlock( AST parent )
	{
		return getASTByType(parent, BLOCK).getText();
	}

    /**
     *  @return the first child of the parent AST with the specified node type.
     */
    private AST getASTByType(AST parent, int node_type)
    {
        for ( AST current = parent.getFirstChild(); current != null; current = current.getNextSibling() )
        {
            if ( current.getType() == node_type )
                return current;
        }

        throw new IllegalStateException ( "AST " + parent.toString() + " has no child of node type " + node_type + "." );

    }

    /**
     *  A JBurgProduction can provide its goal state.
     *  Pattern-matching rules (JBurgRule) and nonterminal
     *  transformation rules (ClosureRecord) both implement
     *  JBurgProduction.
     */
    public interface JBurgProduction
    {
        public abstract String getGoalState();
    }

    /**
     *  JBurgRule contains an AST that represents a rule, 
     *  its associated JBurgReduceAction,
     *  and accessors to the AST's components.
     */
    public class JBurgRule implements JBurgProduction
    {
        /**
         *  The parsed rule.
         */
        AST m_AST;
        
        /**
         *  If the rule is a pattern rule, its pattern matcher.
         */
        JBurgPatternMatcher patternMatcher = null;
        
        /**
         * The rule's reduction action.
         */
        JBurgReduceAction m_reduceAction;

        public JBurgRule(AST n)
        {
            m_AST = n;
            
            if ( m_AST.getType() == PATTERN_RULE )
            {
                AST pattern_root = m_AST.getFirstChild().getNextSibling().getFirstChild();
                try
                {
                    this.patternMatcher = generateMatcher(pattern_root);
                }
                catch ( Exception ex )
                {
                    logger.exception("Building pattern recognizer for " + m_AST.toStringTree(), ex);
                }
            }
        }


        /**
         * @param baseNode - the path to the base of the subtree.
         * @return the subtree's cost specification.
         */
        public String getCost(String baseNode)
        {
            if ( m_AST.getType() != SIMPLE_TRANSFORMATION_RULE )
            {
                AST cost_spec = m_AST.getFirstChild().getNextSibling()
                        .getNextSibling();

                String costText = cost_spec.getFirstChild().getText();

                if ( cost_spec.getType() == SIMPLE_COST_SPEC )
                {
                    return costText;
                }
                else
                {
                    //  TODO: Pass in the active emitter so
                    //  this can be generated as a method call.
                    return costText + "(" + baseNode + ")";
                }
            }
            else
            {
                // Simple transformation rules don't cost anything.
                return "0";
            }
        }

        public JBurgReduceAction getReduceAction()
        {
			if ( null == m_reduceAction )
			{
				throw new IllegalStateException( "getReduceAction() has no action to return." );
			}

            return m_reduceAction;
        }

        public String getOperator()
        {
			int type = m_AST.getType();

			if ( PATTERN_RULE == type )
			{
				return m_AST.getFirstChild().getNextSibling().getFirstChild().getFirstChild().getText();
			}
			else
			{
			    //  A transformation rule.
				return m_AST.getFirstChild().getNextSibling().getText();
			}
        }

        public String getGoalState()
        {
            return m_AST.getFirstChild().getText();
        }

        public void setReduceAction(JBurgReduceAction reduceAction)
        {
            m_reduceAction = reduceAction;
        }

        @Override
        public String toString()
        {
            return m_AST.toStringTree();
        }
    }

    /**
     *  JBurgReduceAction holds action code fragments and their associated "parameters."
     *  A rule specifies "parameters" by naming individual subgoal states within a pattern
     *  rule specifiction, for example,
     *  <xmp>integer = PLUS ( integer i1, integer i2 )</xmp>
     *  In this example, i1 and i2 are the action routine's "parameters."  Since the action
     *  routines all share a common signature, the parameters are passed via the reducer's
     *  reduced values stack.
     */
    private class JBurgReduceAction
    {
        /** The routine's source, from the input specification. */
        private String m_strActionCode;
        
        /**
         * The non-terminal state produced by this action, e.g., expression from
         * <xmp>expression = ADD(expression l, expression r)</xmp>
         */
        private String m_state;
        
        /**
         *  The operator ID of a pattern rule; used to find content-handling
         *  code from the InodeAdapter2.
         */
        private String m_operator;
        
        /**
         * Names and types of the routine's parameters.
         * Types are given as their non-terminal state names; toString()
         * translates these BURM-centric types into the corresponding
         * types in the target language using the mapping set up by
         * the ReturnType directives in the input specification.
         */
        private Vector<ParameterDescriptor> m_parameterList = new Vector<ParameterDescriptor>();

        /**
         *  This action routine's index, assigned in entry order.
         *  The action routines are enumerated as action_1(JBurgNode p), action_2(JBurgNode p),
         *  etc. in the generated reducer.
         */
        int m_index;
        
        /**
         * Track name-to-subtree mappings to be emitted.
         */
        class NamedSubtree
        {
            String path;
            String name;
            
            NamedSubtree(String path, String name)
            {
                this.path = path;
                this.name = name;
            }
        }
        /**
         *  Saved name-to-subtree mappings.
         */
        Vector<NamedSubtree> namedChildNodes = new Vector<NamedSubtree>();
        
        /**
         * Construct a reduce action.
         * @param strActionCode - the action's implementation
         * code, in a target-specific language 
         */
        public JBurgReduceAction(String strActionCode)
        {
            m_strActionCode = strActionCode;
        }

        /**
         * Add a parameter to the action's parameter list.
         */
        public void addParameter(String parmName, String parmState, boolean is_fixed_arity)
            throws Exception
        {
            m_parameterList.add(new ParameterDescriptor(parmName, parmState, is_fixed_arity));
        }

        /**
         *  Return this action's index (the N constituent of its generated name, action_N).
         */
        public int getIndex()
        {
            return m_index;
        }

        /**
         *  Set this action's index (the N constituent of its generated name, action_N).
         */
        public void setIndex(int index)
        {
            m_index = index;
        }

        /**
         * Set the non-terminal state this reduction derives.
         * @param state - the non-terminal state, e.g., expression from
         * <xmp>expression = ADD(expression l, expression r)</xmp>
         */
        public void setState(String state)
        {
            this.m_state = state;
        }

        /**
         * @return the non-terminal state this reduction derives.
         */
        public String getState()
        {
            return this.m_state;
        }

        /**
         *  @param operator the operator at the root of 
         *    the matched subtree.
         */
        public void setOperator(String operator)
        {
            this.m_operator = operator;
        }

        /**
         * @return the operator at the root of 
         *    the matched subtree.
         */
        public String getOperator()
        {
            return m_operator;
        }

        /**
         *  Return the reduction action's code, 
         *  with prepended logic to pop its parameters off the stack.
         */
        @Override
        public String toString()
        {
            String result = "";

            for ( ParameterDescriptor next_param: m_parameterList )
            {
                String paramName = next_param.paramName;
                String paramType = getReturnType(next_param.paramState);

				if ( !next_param.isFixedArity )
				{
					paramType = m_codeEmitter.genNaryContainerType(paramType);
				}
                
                result += "\n" + m_codeEmitter.genPopFromStack(reducedValuesName, paramType, paramName, "\t\t");
            }
            
            for ( NamedSubtree named_child: this.namedChildNodes )
            {
                result += "\n\t" + iNodeClass + " " + named_child.name + " = " + named_child.path + ";";
            }

            // insert 2 tabs into each line, and convert \r\n to \n so the generated
			// files have consistent line endings.
			String[] actionlines = m_strActionCode.toString().split("\n");
			for( int l = 0; l < actionlines.length; ++l )
			{
				if(actionlines[l].endsWith("\r"))
				{
					actionlines[l] = actionlines[l].substring(0,actionlines[l].length()-1);
				}
				result += "\n\t\t" + actionlines[l];
			}

			return result;
        }

        /**
         *  Track a name-to-subtree mapping.  These are unreduced I-nodes,
         *  typically a terminal identified by pattern match.
         *  @param path - the path from the root node to the subtree.
         *  @param name - the name to give to the subtree.
         */
        public void addNamedSubtree(String path, String name)
        {
            namedChildNodes.add(new NamedSubtree(path, name));
        }

    }

    /**
     *  ClosureRecord tracks the target state, cost, and action code associated
     *  with a transitive closure from one state to a target state.
     */
    private class ClosureRecord implements JBurgProduction
    {
        private JBurgRule n;
        private JBurgReduceAction m_reduceAction;

        public ClosureRecord(JBurgRule n) throws Exception
        {
            this.n = n;
			setReduceAction(n.getReduceAction());
        }

        public String getCost(String baseNode)
        {
            return this.n.getCost(baseNode);
        }

        public String getGoalState()
        {
            return this.n.getGoalState();
        }

        public JBurgReduceAction getReduceAction()
        {
            return m_reduceAction;
        }

        public void setReduceAction(JBurgReduceAction reduceAction)
        {

            if (reduceAction != null)
			{
				m_reduceAction = reduceAction;
			}
			else
			{
                throw new IllegalStateException("ClosureRecord cannot accept a null reduce action.");
            }
        }

        /**
         *  Closure records are equal if they have the same goal state, action, and cost.
         *  @see java.lang.Object#equals
         *
         * @param o -- the object to test for equality.
         * @return true if o is a ClosureRecord and it equals this one.
         *
         */
        @Override
        public boolean equals(Object o)
        {
            boolean result = false;

            if ((o != null) && o instanceof ClosureRecord) {
                ClosureRecord cuz = (ClosureRecord) o;

                result = getGoalState().equals(cuz.getGoalState());
                result &= m_reduceAction.equals(cuz.m_reduceAction);
                result &= getCost(reducerNodeName).equals(cuz.getCost(reducerNodeName));
            }

            return result;
        }
    }

	/**
	 *  A ParameterDescriptor is a pair (name, subgoalState) which parallels
	 *  the FOO(name, subgoalState) found in the specification's syntax.
	 */
    private static class ParameterDescriptor
    {
        String paramName;
        String paramState;
        boolean isFixedArity = true;
        
        final static boolean FIXED_ARITY = true;

        ParameterDescriptor(String paramName, String paramState, boolean is_fixed_arity)
        {
            this.paramName = paramName;
            this.paramState = paramState;
            this.isFixedArity = is_fixed_arity;
        }
    }


	private void emitComputeCostMatrixFunction(Map<String, Vector<JBurgRule>> patternRules, String iNodeClass, PrintStream output) throws Exception 
	{
		//  Emit the method declaration and local variables.
		String decl = m_codeEmitter.declareMethod(
			Modifier.PRIVATE,
			"void",
			"computeCostMatrix",
			new String[][]{
				new String[] { "JBurgAnnotation", "node" }
			},
			null
		);

		output.print( decl );

		output.print( m_codeEmitter.genBeginBlock() ); 
		output.print( m_codeEmitter.genLocalVar( "int", "iCost", null ) );

		//  The calculations are based on finding patterns
		//  that can label the node's type.
		output.print( 
			m_codeEmitter.genBeginEnumeratedChoiceSet( 
				m_codeEmitter.genCallMethod(
					"node",
					"getOperator",
					null
				)
			)
		);

		//  Try each pattern for a match.
		//  Any pattern that matches then checks its 
		//  cost to see if it's less than the current
		//  best cost for the node, and replaces the 
		//  node's current rule and cost with its own
		//  if it's the new best cost.

		for (Vector<JBurgRule> vPatternRules: patternRules.values())
		{
			//  Emit all the patterns that match the current node type.
			output.print(
				m_codeEmitter.genBeginChoiceCase(
					vPatternRules.firstElement().getOperator()
				)
			);

			for (JBurgGenerator.JBurgRule nPatternRule: vPatternRules)
			{
				try
				{
					emitPatternRule(nPatternRule, iNodeClass, output);
				}
				catch ( Exception ex )
				{
                    getLogger().exception(
                        "processing pattern: \"" + nPatternRule.m_AST.getFirstChild().getNextSibling().toStringList() + "\"  goal state " + nPatternRule.getGoalState(),
                        ex
                    );

					m_errCount++;
				}
			}

			output.print( m_codeEmitter.genEndChoiceCase() );
		}

		output.print( m_codeEmitter.genEndEnumeratedChoiceSet() );
		output.print( m_codeEmitter.genEndBlock() );
	}
	
	/**
	 *  Emit a routine that inspects the operators found at compiler compile time,
	 *  and emits code to get the corresponding node's Nth child.
	 *  @param output
	 */
	private void emitGetNthChild(PrintStream output)
	{
	    String decl = m_codeEmitter.declareMethod(
	            Modifier.PRIVATE,
	            this.iNodeClass,
	            "getNthChild",
	            new String[][]{
	                { this.iNodeClass, "node" },
	                { "int", "index" }
	            },
	            null
	        );
	    output.print(decl);
	    output.print(this.m_codeEmitter.genBeginBlock());
	    output.print(this.m_codeEmitter.genLocalVar(this.iNodeClass, "result", null));
	    
	    output.print(m_codeEmitter.genBeginEnumeratedChoiceSet(this.m_adapter.genGetOperator("node", this.m_codeEmitter)));
	    for ( String opcode: this.allOperators)
	    {
	        Integer choice_count = this.adapter2.getMaxNthChildChoice(opcode);
	        if ( choice_count != null )
	        {
	            output.print(m_codeEmitter.genBeginChoiceCase(opcode));
	            output.print(m_codeEmitter.genBeginEnumeratedChoiceSet("index"));
	            for ( int i = 0; i < choice_count; i++ )
	            {
	                output.print(m_codeEmitter.genBeginChoiceCase(Integer.toString(i)));
	                output.print(this.m_codeEmitter.genAssignment("result", this.adapter2.genGetNthChild(opcode, "node", i, m_codeEmitter)));
	                output.print(this.m_codeEmitter.genEndChoiceCase());
	                
	            }
	            output.print(this.m_codeEmitter.genDefaultChoiceCase());
	            output.print(m_codeEmitter.genAssignment("result", this.adapter2.genGetDefaultChild(opcode, "node", "index", m_codeEmitter)));
	            output.print(m_codeEmitter.genEndEnumeratedChoiceSet());
	            output.print(this.m_codeEmitter.genEndChoiceCase());
	        }
	        
	    }
	    output.print(this.m_codeEmitter.genDefaultChoiceCase());
	    output.print(m_codeEmitter.genAssignment("result", this.m_adapter.genGetNthChild("node", "index", m_codeEmitter)));
	    output.print(m_codeEmitter.genEndEnumeratedChoiceSet());
	    output.print(m_codeEmitter.genBeginLine() + m_codeEmitter.genReturnValue("result") + m_codeEmitter.genEndStmt());
	    output.print(this.m_codeEmitter.genEndBlock());
	}

	private void emitPatternRule(JBurgGenerator.JBurgRule p, String iNodeClass, PrintStream output) throws Exception
	{
		AST pattern_root = p.m_AST.getFirstChild().getNextSibling().getFirstChild();

		

		String encoded_pattern = p.patternMatcher.generatePatternRecognizer(m_codeEmitter, "node", this.adapter2);

		//  The pattern may have been optimized out.
		if ( encoded_pattern != null )
		{
    		//  Emit the pattern match.
            output.print( m_codeEmitter.genBeginLine() );
    		output.print( m_codeEmitter.genIf (encoded_pattern) );
    		output.print( m_codeEmitter.genBeginBlock() );
		}

        /*
		 *   Emit code to compute this rule's cost, which is the cost of the rule itself 
		 *   and all the non-terminals associated with it; then emit a comparison with
		 *   the current best cost.
		 */

		//  Compute the basic cost of the pattern match.
		String match_cost = p.getCost( 
			m_codeEmitter.genCast( 
				iNodeClass, 
				m_codeEmitter.genAccessMember("node", "m_node")
			)
		);

		String sub_costs = null;
		
		for ( JBurgPatternMatcher subgoal: p.patternMatcher.getParameterizedSubtrees() )
		{
		    String subgoal_cost = subgoal.generateCost(m_codeEmitter, "node");
		    if ( null == sub_costs )
		    {
		        sub_costs = subgoal_cost;
		    }
		    else if ( subgoal_cost != null )
		    {
		        sub_costs = m_codeEmitter.genAddition(sub_costs, subgoal_cost);
		    }
		}
		
		if ( sub_costs != null && sub_costs.length() > 0 )
		    match_cost = m_codeEmitter.genAddition(match_cost, sub_costs);
		
		output.print(m_codeEmitter.genBeginLine() + m_codeEmitter.genAssignment("iCost", match_cost));

		//  Does this pattern match cost less than the previously best match?
		String strCostCall = m_codeEmitter.genCallMethod ( "node", "getCost", new String[] {  m_codeEmitter.genGetGoalState(p) } );

		output.print( m_codeEmitter.genIf( m_codeEmitter.genCmpLess( strCostCall, "iCost" )));
		output.print( m_codeEmitter.genBeginBlock() );

		//  Emit code to handle a successful match.
		output.print( m_codeEmitter.genBeginLine() );
		output.print( m_codeEmitter.genComment( "Matched " + p.getOperator() + " ==> " + p.getGoalState()));
		output.print( m_codeEmitter.genBeginLine() );

		//  Reset the reduce-time rule to fire.
        String strRule = String.valueOf(p.getReduceAction().getIndex());

        output.print( 
			m_codeEmitter.genCallMethod(
				"node", 
				"reset", 
				new String[] { m_codeEmitter.genGetGoalState(p), "iCost", strRule }
			)
		);

		output.print( m_codeEmitter.genEndStmt() );

		if ( p.patternMatcher.isNary() )
		{
			output.print( 
				m_codeEmitter.genCallMethod(
					"node", 
					"setNarySubgoal", 
					new String[] { m_codeEmitter.genGetGoalState(p), m_codeEmitter.genGetGoalState(p.patternMatcher.getSubgoal() ) }
				)
			);

			output.print( m_codeEmitter.genEndStmt() );
		}
		else
		{
			//  Set up subgoals for the operands' reductions.
			for ( JBurgPatternMatcher subgoal : p.patternMatcher.getParameterizedSubtrees() )
			{
				
			    if ( ! subgoal.isNary() )
			    {
    				output.print (
    					m_codeEmitter.genCallMethod (
    						"node",
    						"addSubgoal",
    						new String[] 
    						{  
    							m_codeEmitter.genGetGoalState(p),  
    							subgoal.generatePathToRoot(m_codeEmitter, "node"),
    							m_codeEmitter.genGetGoalState(subgoal.getSubgoal() )
    						}
    					)
    				);
			    }
			    else
			    {
			        output.print( 
			                m_codeEmitter.genCallMethod(
			                    "node",
			                    "addNarySubgoal", 
			                    new String[] 
			                    {
			                        m_codeEmitter.genGetGoalState(p), 
			                        subgoal.generatePathToRoot(m_codeEmitter, "node"),
			                        m_codeEmitter.genGetGoalState(subgoal.getSubgoal()),
			                        subgoal.m_positionInParent.toString()
			                    }
			                )
			            );
			    }

				output.print( m_codeEmitter.genEndStmt() );
			}
        }


        if (hasClosure(p))
		{
            output.print( 
				m_codeEmitter.genCallMethod( 
					null, 
					"closure_" + p.getGoalState(),
					new String[] { "node", "iCost" }
				)
			);

			output.print( m_codeEmitter.genEndStmt() );
        }

        if ( encoded_pattern != null)
            output.print( m_codeEmitter.genEndBlock() );
        output.print( m_codeEmitter.genEndBlock() );

        //  Add the non-terminal subgoal result nodes to the action routine's "parameters."
        //  In actuality, these values are popped off the reduce action stack
        //  during the action routine's setup phase.
        JBurgGenerator.JBurgReduceAction action = p.getReduceAction();

        for (JBurgPatternMatcher subgoal: p.patternMatcher.getParameterizedSubtrees())
		{
            action.addParameter(subgoal.getParameterName(), subgoal.getSubgoal(), !subgoal.isNary());
        }
        
        //  Add the named subtrees as a convenience.
        for ( JBurgPatternMatcher named_terminal: p.patternMatcher.getNamedSubtrees())
        {
            action.addNamedSubtree(named_terminal.generatePathToRoot(m_codeEmitter, reducerNodeName), named_terminal.getParameterName());
        }
        
        //  Send the root node's operator to the action routine,
        //  so that the action-code emitter can substitute in
        //  content-handling code.
        if ( p.patternMatcher.matchesOperator() )
            action.setOperator(p.patternMatcher.getOperator());
	}
	
	private JBurgPatternMatcher generateMatcher(AST pattern_root)
	throws Exception
	{   
        JBurgPatternEncoder patternBURM = new JBurgPatternEncoder();

        //  As we traverse the subtree, we may find parameterized subtrees,
        //  for example, in ADD(expr lhs, expr rhs) lhs and rhs are paramterized
        //  subtrees.  These subtrees play several parts in the computation of the
        //  locally-optimal reduction:
        //  -  They contribute to the rule's computed cost.
        //  -  The reduction's action code may refer to these elements by name.
        //  -  If the rule is of the form OP(nttype1 a [, nttype2 b]), then the rule
        //     must enforce this reduce-time goal state on its subtrees' reductions.
        patternBURM.setSubgoals(new Vector<JBurgPatternMatcher>());
        
        //  There may also be named terminals in the pattern;
        //  as a convenience, record these so that they
        //  can be used by name in the reduction.
        patternBURM.setNamedterminals(new Vector<JBurgPatternMatcher>());
        
        patternBURM.setOperators(this.allOperators);

        try
        {
            patternBURM.burm( pattern_root );
        }
        catch (IllegalStateException burm_error )
        {
            if ( this.patternMatcherDumpFile != null )
            {
                //  Dump the BURM's debugging info.
                java.io.PrintWriter dumper = new java.io.PrintWriter(new java.io.FileWriter(patternMatcherDumpFile));
                dumper.println ( "<?xml version=\"1.0\"?>");
                dumper.println("<BurmDump date=\"" + new java.util.Date().toString() + "\">");
                patternBURM.dump(dumper);
                dumper.println("<AST>");
                dumper.println("<![CDATA[");
                dumper.println(pattern_root.toStringTree());
                dumper.println("]]>");
                dumper.println("</AST>");
                dumper.println("</BurmDump>");
                dumper.flush();
                dumper.close();
            }
            
            throw burm_error;
        }
        
        JBurgPatternMatcher recognizer = (JBurgPatternMatcher) patternBURM.getResult();
        recognizer.setParameterizedSubtrees(patternBURM.getSubgoals());
        recognizer.setNamedSubtrees(patternBURM.getNamedterminals());
        
        return recognizer;
	}

	private void emitActions(Vector<JBurgReduceAction> reduceActions, String iNodeClass, PrintStream output)
	{
		Iterator<JBurgReduceAction> it = reduceActions.iterator();
		int i;

		String[][] actionParameterDescriptor = new String[][] { new String[] { iNodeClass, reducerNodeName } };

		//  Print out the individual action routines.
		//  Action #1 is the builtin trivial action for simple transformation rules.
		for (i = 2, it.next(); it.hasNext(); i++)
		{
			JBurgGenerator.JBurgReduceAction nextAction = it.next();

			output.print(m_codeEmitter.genBeginLine());
			output.print(m_codeEmitter.genComment(nextAction.getState()));
			
			String decl = m_codeEmitter.declareMethod(
				Modifier.PRIVATE,
				getReturnType(nextAction.getState()),
				"action_" + String.valueOf(i),
				actionParameterDescriptor,
				null
			);
				
			output.print( decl );
			output.print( m_codeEmitter.genBeginBlock() );

			String action_routine = nextAction.toString();
			
			//  Replace #OPERATOR# with a content-access snippet.
			if ( this.adapter2 != null && nextAction.getOperator() != null )
			{
			    String content_access = this.adapter2.genGetContent(nextAction.getOperator(), this.reducerNodeName, nextAction.getState(), m_codeEmitter);
			    
			    if ( content_access != null )
			    {
			        String content_pattern = "#" + nextAction.getOperator() + "#";
			        action_routine = action_routine.replaceAll( content_pattern, content_access );
			    }
			}
			//  Replace #goalstate with the name of the action's input node.
			
			String goalPattern = "#" + nextAction.getState();
			action_routine = action_routine.replaceAll( goalPattern, this.reducerNodeName );
			
			output.print( action_routine );

			output.print( m_codeEmitter.genEndBlock() );
		}

		//  Emit their common dispatch routine.
		String dispatchDecl = m_codeEmitter.declareMethod(
			Modifier.PRIVATE,
			"void",
			"dispatchAction",
			new String[][]{
				new String[] { iNodeClass, this.reducerNodeName },
				new String[] { "int", "iRule" }
			},
			new Class[] { Exception.class }
		);

		output.print( dispatchDecl );
		output.print( m_codeEmitter.genBeginBlock() );
		output.print( m_codeEmitter.genBeginEnumeratedChoiceSet( "iRule" ) );

		output.print( m_codeEmitter.genBeginChoiceCase( "1" ) );
		output.print( m_codeEmitter.genComment( "Don't reduce or touch the stack." ) );
		output.print( m_codeEmitter.genBeginLine() );
		output.print( m_codeEmitter.genEndChoiceCase() );

		//  Emit the dispatch case statement.
		//  Action #1 is the builtin trivial action for simple transformation rules,
		//  and the dispatcher knows that it can be elided, so this loop starts with
		//  action routine 2.
		for (i = 2; i <= reduceActions.size(); i++)
		{
			output.print( m_codeEmitter.genBeginChoiceCase( String.valueOf(i)));
			output.print( 
				m_codeEmitter.genPushToStack(
					reducedValuesName,
					m_codeEmitter.genCallMethod(
						"this",
						"action_" + String.valueOf(i),
						new String[] { reducerNodeName }
					)
				)
			);
			output.print( m_codeEmitter.genEndStmt() );
			output.print( m_codeEmitter.genEndChoiceCase() );
		}

		output.print( m_codeEmitter.genDefaultChoiceCase() );
		output.print( m_codeEmitter.genSignalError ( "\"Unmatched reduce action \" + iRule" ) );
		output.print( m_codeEmitter.genEndStmt() );
	    
		output.print ( m_codeEmitter.genEndBlock() );
		output.print ( m_codeEmitter.genEndBlock() );
	}

	private void emitNTConstants(Set<String> goal_states, PrintStream output)
	{
		Iterator<String> keys = goal_states.iterator();

		int nthConstant = 0;

		output.print(m_codeEmitter.genBeginLine());
		while (keys.hasNext())
		{
			nthConstant++;

			String strNTName = m_codeEmitter.genGetGoalState(keys.next().toString() );

			output.print( 
				m_codeEmitter.genInstanceField(
					Modifier.PUBLIC + Modifier.STATIC + Modifier.FINAL,
					"int",
					strNTName,
					String.valueOf(nthConstant)
				)
			);
		}
	
		output.print(
			m_codeEmitter.genInstanceField(
				Modifier.PUBLIC + Modifier.STATIC + Modifier.FINAL,
				"int",
				"nStates",
				String.valueOf(nthConstant)
			)
		);
	}

	public void emitCostFns(Vector<AST> cost_functions, String iNodeClass, PrintStream output) 
	{
		String[][] costParms = new String[][] {
			new String[] { iNodeClass, reducerNodeName }
		};

		for (int i = 0; i < cost_functions.size(); i++)
		{
			AST currentNode = cost_functions.elementAt(i);

			String functionName = currentNode.getFirstChild().getText();

			output.print( m_codeEmitter.declareMethod ( Modifier.PRIVATE, "int", functionName, costParms, null ));
			output.print( m_codeEmitter.genBeginLine() );
			output.print( getCodeBlock(currentNode) );
		}
	}

	public void emitClosures(Map<String, Vector<ClosureRecord>> closureSets, String iNodeClass, PrintStream output)
	{
		for (String strClosureNT: closureSets.keySet())
		{
			output.print(
				m_codeEmitter.declareMethod(
					Modifier.PRIVATE,
					"void",
					"closure_" + strClosureNT,
					new String[][]{
						new String[] { "JBurgAnnotation", reducerNodeName },
						new String[] { "int", "c" }
					},
					null
				)
			);

			output.print ( m_codeEmitter.genBeginBlock() );

			output.print( m_codeEmitter.genLocalVar ("int", "iCost", null ));

			for (JBurgGenerator.ClosureRecord newClosure: closureSets.get(strClosureNT) )
			{
				String strRule;

				if (newClosure.getReduceAction() != null)
					strRule = String.valueOf(newClosure.getReduceAction()
													   .getIndex());
				else
					strRule = "0";

				String closureCost = newClosure.getCost( m_codeEmitter.genAccessMember(reducerNodeName, "m_node"));

				if (closureCost.equals("0") )
				{
					genAssignment( output, "iCost", "c");
				}
				else
				{
					genAssignment( output, "iCost",  m_codeEmitter.genAddition( "c", closureCost ));
				}

				output.print(
					m_codeEmitter.genIf(
						m_codeEmitter.genCmpLess(
							m_codeEmitter.genCallMethod ( reducerNodeName, "getCost", new String[] { m_codeEmitter.genGetGoalState(newClosure.getGoalState()) } ),
							"iCost"
						)
					)
				);
				output.print( m_codeEmitter.genBeginBlock() );

				output.print(
					m_codeEmitter.genCallMethod(
						reducerNodeName,
						"reset",
						new String[] { m_codeEmitter.genGetGoalState(newClosure.getGoalState()), "iCost", strRule }
					)
				);
				output.print( m_codeEmitter.genEndStmt() );

				output.print(
					m_codeEmitter.genCallMethod(
						reducerNodeName,
						"recordAntecedent",
						new String[] { m_codeEmitter.genGetGoalState(newClosure.getGoalState() ), m_codeEmitter.genGetGoalState(strClosureNT ) }
					)
				);
				output.print( m_codeEmitter.genEndStmt() );

				if (hasClosure(newClosure))
				{
					output.print(
						m_codeEmitter.genCallMethod(
							"this",
							"closure_" + newClosure.getGoalState(),
							new String[] { reducerNodeName, "iCost" }
						)
					);
					output.print( m_codeEmitter.genEndStmt() );
				}

				output.print( m_codeEmitter.genEndBlock() );
			}

			output.print( m_codeEmitter.genEndBlock() );
		}
	}


	public void emitLabelFunction(String iNodeClass, PrintStream output)
	{
		output.print(
			m_codeEmitter.declareMethod(
				Modifier.PUBLIC,
				"JBurgAnnotation",
				"label",  
				new String[][] { new String[] { iNodeClass, m_initalParamName } },
				null
			)
		);

		output.print( m_codeEmitter.genBeginBlock() );
		output.print( m_codeEmitter.genLocalVar ( "JBurgAnnotation", "result", m_codeEmitter.genNullPointer() ) );
		output.print( m_codeEmitter.genLocalVar ( "int", "i", null) );
		output.print( m_codeEmitter.genLocalVar ( "int", "arity", null) );

		//  if (to_be_labelled != null )
		output.print( m_codeEmitter.genIf ( m_codeEmitter.genCheckPtr( m_initalParamName, false )));
		output.print( m_codeEmitter.genBeginBlock() );

		genAssignment( 
			output, 
			"result", 
			m_codeEmitter.genNewObject(
				"JBurgAnnotation", 
				new String[] { m_initalParamName, "nStates + 1" } 
			)
		);

		//  Generate a loop over children that labels them,
		//  and adds their labelled JBurgAnnotation nodes to the current node.
		genAssignment(output, "arity", m_adapter.genGetArity(m_initalParamName, m_codeEmitter));	
		genAssignment(output, "i", "0");

		output.print(m_codeEmitter.genPreTestLoop( m_codeEmitter.genCmpLess("arity", "i")));
		output.print(m_codeEmitter.genBeginBlock());

		// An Inode​Adapter2 capable BURM needs to call the routine that uses the
		// per-operator code-gen in Inode​Adapter2 to call the correct routines to
		// get the Nth child.
		String nthChildAccessor;
		
		if (this.adapter2 != null) 
		{
		    nthChildAccessor = m_codeEmitter.genCallMethod("this", "getNthChild", new String[] { m_initalParamName, "i" });
		}
		else
	    {
	        nthChildAccessor = m_adapter.genGetNthChild( m_initalParamName, "i", m_codeEmitter );
	    }
		
		String [] nthChildParam = new String[] {
			m_codeEmitter.genCallMethod(
				"this",
				"label",
				new String[] {
					m_codeEmitter.genCast( 
						iNodeClass, 
						nthChildAccessor
					)
				}
			)
		};

		output.print( m_codeEmitter.genBeginLine() );
		output.print( m_codeEmitter.genCallMethod( "result", "addChild", nthChildParam ));
		output.print( m_codeEmitter.genEndStmt());
		
		genAssignment(output, "i", m_codeEmitter.genAddition("i", "1") );

		output.print(m_codeEmitter.genEndBlock()); // while

		//  Call the costing function.
		output.print( m_codeEmitter.genCallMethod( "this", "computeCostMatrix", new String[] { "result" } ) );
		output.print( m_codeEmitter.genEndStmt() );

		output.print( m_codeEmitter.genEndBlock() );  // if to_be_labelled != null

		output.print( m_codeEmitter.genReturnValue( "result" ) );
		output.print( m_codeEmitter.genEndStmt() );

		output.print( m_codeEmitter.genEndBlock() );
	}

	void genAssignment(PrintStream output, String lvalue, String rvalue)
	{
		output.print( m_codeEmitter.genAssignment( lvalue, rvalue ) );
	}

    Logger getLogger()
    {
        if ( null == this.logger )
            this.logger = new Logger(true, true, true);
        return this.logger;
    }
}
